@@ -9,12 +9,11 @@
  *
  */
 
-#ifndef __CLIENT_OPTIONS_H__
-#define __CLIENT_OPTIONS_H__
+#pragma once
 
 typedef struct memcached_help_text_st memcached_help_text_st;
 
-typedef enum {
+enum memcached_options {
   OPT_SERVERS= 's',
   OPT_VERSION= 'V',
   OPT_HELP= 'h',
@@ -36,8 +35,10 @@ typedef enum {
   OPT_HASH,
   OPT_BINARY,
   OPT_UDP,
+  OPT_BUFFER,
   OPT_USERNAME,
-  OPT_PASSWD
-} memcached_options;
-
-#endif /* CLIENT_OPTIONS */
+  OPT_PASSWD,
+  OPT_STAT_ARGS,
+  OPT_QUIET,
+  OPT_FILE= 'f'
+};
