@@ -63,16 +63,27 @@
 #include <sys/types.h>
 #include <unistd.h>
 
-#include <libmemcached/memcached.h>
+#ifdef HAVE_STRINGS_H
+#include <strings.h>
+#endif
+
+#include <libmemcached-1.0/memcached.h>
 #include <libmemcached/watchpoint.h>
 #include <libmemcached/is.h>
 
+#include <libmemcached/server_instance.h>
+
+#ifdef HAVE_POLL_H
+#include <poll.h>
+#else
+#include "poll/poll.h"
+#endif
+
+
 #ifdef __cplusplus
 extern "C" {
 #endif
 
-typedef struct memcached_server_st * memcached_server_write_instance_st;
-
 typedef memcached_return_t (*memcached_server_execute_fn)(memcached_st *ptr, memcached_server_write_instance_st server, void *context);
 
 LIBMEMCACHED_LOCAL
@@ -91,22 +102,41 @@ memcached_return_t memcached_server_execute(memcached_st *ptr,
 #include <libmemcached/error.hpp>
 #include <libmemcached/memory.h>
 #include <libmemcached/io.h>
+#ifdef __cplusplus
+#include <libmemcached/string.hpp>
+#include <libmemcached/io.hpp>
+#include <libmemcached/udp.hpp>
 #include <libmemcached/do.hpp>
+#include <libmemcached/close_socket.hpp>
+#include <libmemcached/connect.hpp>
+#include <libmemcached/allocators.hpp>
+#include <libmemcached/hash.hpp>
+#include <libmemcached/quit.hpp>
+#include <libmemcached/server.hpp>
+#include <libmemcached/behavior.hpp>
+#include <libmemcached/sasl.hpp>
+#include <libmemcached/server_list.hpp>
+#endif
 #include <libmemcached/internal.h>
 #include <libmemcached/array.h>
 #include <libmemcached/libmemcached_probes.h>
 #include <libmemcached/memcached/protocol_binary.h>
 #include <libmemcached/byteorder.h>
 #include <libmemcached/initialize_query.h>
+#ifdef __cplusplus
 #include <libmemcached/response.h>
-#include <libmemcached/prefix_key.h>
+#endif
+#include <libmemcached/namespace.h>
+#include <libmemcached/virtual_bucket.h>
 
-/* string value */
-struct memcached_continuum_item_st
-{
-  uint32_t index;
-  uint32_t value;
-};
+#ifdef __cplusplus
+#include <libmemcached/backtrace.hpp>
+#include <libmemcached/assert.hpp>
+#include <libmemcached/server.hpp>
+#include <libmemcached/key.hpp>
+#endif
+
+#include <libmemcached/continuum.hpp>
 
 #if !defined(__GNUC__) || (__GNUC__ == 2 && __GNUC_MINOR__ < 96)
 
@@ -128,9 +158,6 @@ struct memcached_continuum_item_st
 extern "C" {
 #endif
 
-LIBMEMCACHED_LOCAL
-memcached_return_t memcached_connect(memcached_server_write_instance_st ptr);
-
 LIBMEMCACHED_LOCAL
 memcached_return_t run_distribution(memcached_st *ptr);
 
@@ -138,45 +165,9 @@ memcached_return_t run_distribution(memcached_st *ptr);
 #define memcached_server_response_decrement(A) (A)->cursor_active--
 #define memcached_server_response_reset(A) (A)->cursor_active=0
 
-LIBMEMCACHED_LOCAL
-void set_last_disconnected_host(memcached_server_write_instance_st ptr);
-
-LIBMEMCACHED_LOCAL
-memcached_return_t memcached_key_test(const char * const *keys,
-                                      const size_t *key_length,
-                                      size_t number_of_keys);
-
 LIBMEMCACHED_LOCAL
 memcached_return_t memcached_purge(memcached_server_write_instance_st ptr);
 
-LIBMEMCACHED_LOCAL
-memcached_server_st *memcached_server_create_with(const memcached_st *memc,
-                                                  memcached_server_write_instance_st host,
-                                                  const char *hostname,
-                                                  in_port_t port,
-                                                  uint32_t weight,
-                                                  memcached_connection_t type);
-
-
-static inline memcached_return_t memcached_validate_key_length(size_t key_length, bool binary)
-{
-  unlikely (key_length == 0)
-    return MEMCACHED_BAD_KEY_PROVIDED;
-
-  if (binary)
-  {
-    unlikely (key_length > 0xffff)
-      return MEMCACHED_BAD_KEY_PROVIDED;
-  }
-  else
-  {
-    unlikely (key_length >= MEMCACHED_MAX_KEY)
-      return MEMCACHED_BAD_KEY_PROVIDED;
-  }
-
-  return MEMCACHED_SUCCESS;
-}
-
 #ifdef __cplusplus
 }
 #endif
